
// Conrad DCF-Modul Pin4 auf AVR PD2
// ATmega8, 8MHz
// Optimierungsgrad: O2
// erstellt am 08/12/2010

#include "dcf77.h"

unsigned char neu_min      = 0;		//unsigned char: 0-255 !!
unsigned char signal_count = 0;
unsigned char hilfs_count  = 0;
unsigned char sek_count    = 0;
unsigned char sicher_count = 0;
unsigned char parity	   = 0;

//angezeigte Zeit
unsigned char dcf77_jah = 0;  
unsigned char dcf77_mon = 0;
unsigned char dcf77_tag = 0;
unsigned char dcf77_std = 0;
unsigned char dcf77_min = 0;
unsigned char dcf77_sek = 0;

//von DCF77 empfangene Zeit
unsigned char dcf77_jah_h = 0;
unsigned char dcf77_mon_h = 0;
unsigned char dcf77_tag_h = 0;
unsigned char dcf77_std_h = 0;
unsigned char dcf77_min_h = 0;
unsigned char dcf77_sek_h = 0;

//Ausgabe als String
char dcf77_zeitausg[10]  = {'\0'};
char dcf77_datumausg[10] = {'\0'};



// --------- Inizialisierung --------

void dcf77_init (void) {

	//Timer init
	TCCR0 = (1<<CS02) | (0<<CS01) | (0<<CS00);   //Setzen des Prescaler auf 256
	TIMSK |= (1<<TOIE0);
	// Interrupt Aktion alle ( 8000000/256) / 256 = 122,07 Hz -> 0,008192s
	// Interrupt Frequenz >= 100 Hz, sonst wird 'signal_count' zu ungenau!

	//Eingangspin
	DDRD  &= ~(1<<PD2);    // Pin PD2 als Eingang
	PORTD |=  (1<<PD2);    // internen Pull-Up an PD2 aktivieren
}


// --------- Time Overflow alle 0,008192s --------
// muss in der main.c ausgefhrt werden: ISR(TIMER0_OVF_vect){ ... }

void dcf77_TIMER (void) {

	neu_min++;
	sek_count++;

	if (!(PIND & (1<<PD2))){    //bei low Pegel
		signal_count++;			//zhlt das Signal bis max 200ms / 0,008192s = 25
		neu_min = 0;
	}
	if (signal_count > 1){		//fngt mit dem Signal an zu zhlen
		hilfs_count++;
	}
	if (hilfs_count == 60){		//Signal muss fertig sein bei 50 * 0,008192s = 0,5s
		dcf77_bit();
		dcf77_sek_h++;
		hilfs_count  = 0;		//Hilfszhler wird zurckgesetzt
		signal_count = 0;		//Zhler wird zurckgesetzt		
	}

	if (neu_min == 160){		//beginn einer neuen Minute, es wird kein Signal gesendet
								//160 * 0,008192s = 1,31s > 1s
		neu_min 	 = 0;
		dcf77_sek_h  = 0;
		hilfs_count  = 0;
		dcf77_exec();
	}
	
	if (sek_count == 122){		//122 * 0,008192s = 1s  Sekundenzhler
								//damit die Uhr auch ohne Signal luft
		sek_count = 0;
		add_sek();
	}
}


// --------- Auswertung der Bits --------

void dcf77_bit (void) {

	if (signal_count > 20){		//20 * 0,008192s = 160ms  -> Bit=1

		if (dcf77_sek_h == 20) sicher_count++;	//berprfung Bit20 = 1

		//Zeit
		if (dcf77_sek_h == 21) { dcf77_min_h = dcf77_min_h +  1; parity++; }
		if (dcf77_sek_h == 22) { dcf77_min_h = dcf77_min_h +  2; parity++; }
		if (dcf77_sek_h == 23) { dcf77_min_h = dcf77_min_h +  4; parity++; }
		if (dcf77_sek_h == 24) { dcf77_min_h = dcf77_min_h +  8; parity++; }
		if (dcf77_sek_h == 25) { dcf77_min_h = dcf77_min_h + 10; parity++; }
		if (dcf77_sek_h == 26) { dcf77_min_h = dcf77_min_h + 20; parity++; }
		if (dcf77_sek_h == 27) { dcf77_min_h = dcf77_min_h + 40; parity++; }

		if (dcf77_sek_h == 29) { dcf77_std_h = dcf77_std_h +  1; parity++; }
		if (dcf77_sek_h == 30) { dcf77_std_h = dcf77_std_h +  2; parity++; }
		if (dcf77_sek_h == 31) { dcf77_std_h = dcf77_std_h +  4; parity++; }
		if (dcf77_sek_h == 32) { dcf77_std_h = dcf77_std_h +  8; parity++; }
		if (dcf77_sek_h == 33) { dcf77_std_h = dcf77_std_h + 10; parity++; }
		if (dcf77_sek_h == 34) { dcf77_std_h = dcf77_std_h + 20; parity++; }
		//if (dcf77_sek_h == 17) dcf77_std_h = dcf77_std_h +  1; //Sommerzeit

		//Datum
		if (dcf77_sek_h == 36) { dcf77_tag_h = dcf77_tag_h +  1; parity++; }
		if (dcf77_sek_h == 37) { dcf77_tag_h = dcf77_tag_h +  2; parity++; }
		if (dcf77_sek_h == 38) { dcf77_tag_h = dcf77_tag_h +  4; parity++; }
		if (dcf77_sek_h == 39) { dcf77_tag_h = dcf77_tag_h +  8; parity++; }
		if (dcf77_sek_h == 40) { dcf77_tag_h = dcf77_tag_h + 10; parity++; }
		if (dcf77_sek_h == 41) { dcf77_tag_h = dcf77_tag_h + 20; parity++; }

		if (dcf77_sek_h == 42) parity++;  //Wochentage
		if (dcf77_sek_h == 43) parity++;
		if (dcf77_sek_h == 44) parity++;

		if (dcf77_sek_h == 45) { dcf77_mon_h = dcf77_mon_h +  1; parity++; }
		if (dcf77_sek_h == 46) { dcf77_mon_h = dcf77_mon_h +  2; parity++; }
		if (dcf77_sek_h == 47) { dcf77_mon_h = dcf77_mon_h +  4; parity++; }
		if (dcf77_sek_h == 48) { dcf77_mon_h = dcf77_mon_h +  8; parity++; }
		if (dcf77_sek_h == 49) { dcf77_mon_h = dcf77_mon_h + 10; parity++; }

		if (dcf77_sek_h == 50) { dcf77_jah_h = dcf77_jah_h +  1; parity++; }
		if (dcf77_sek_h == 51) { dcf77_jah_h = dcf77_jah_h +  2; parity++; }
		if (dcf77_sek_h == 52) { dcf77_jah_h = dcf77_jah_h +  4; parity++; }
		if (dcf77_sek_h == 53) { dcf77_jah_h = dcf77_jah_h +  8; parity++; }
		if (dcf77_sek_h == 54) { dcf77_jah_h = dcf77_jah_h + 10; parity++; }
		if (dcf77_sek_h == 55) { dcf77_jah_h = dcf77_jah_h + 20; parity++; }
		if (dcf77_sek_h == 56) { dcf77_jah_h = dcf77_jah_h + 30; parity++; }
		if (dcf77_sek_h == 57) { dcf77_jah_h = dcf77_jah_h + 40; parity++; }
	}

	if (dcf77_sek_h ==  0 && signal_count < 15) sicher_count++;	//berprfung Bit0 = 0

	//Paritten (gerade Paritten fr Minute, Stunde und Datum)
	//			(parity gerade -> Bit = 0 ; parity ungerade -> Bit = 1)
	//zustzlich berprfung von max-Wert fr Min, Std, Tag und Monat
	if (dcf77_sek_h == 28 && dcf77_min_h < 61) {
		if (parity%2 == 0 && signal_count < 15) sicher_count++;
		if (parity%2 != 0 && signal_count > 20) sicher_count++;
		parity = 0;	
	}
	if (dcf77_sek_h == 35 && dcf77_std_h < 25) {
		if (parity%2 == 0 && signal_count < 15) sicher_count++;
		if (parity%2 != 0 && signal_count > 20) sicher_count++;
		parity = 0;	
	}
	if (dcf77_sek_h == 58 && dcf77_tag_h < 32 && dcf77_mon_h < 13) {
		sicher_count++;
		if (parity%2 == 0 && signal_count < 15) sicher_count++;
		if (parity%2 != 0 && signal_count > 20) sicher_count++;
		parity = 0;	
	}
	if (dcf77_sek_h == 59) sicher_count--;  //nur wenn GENAU 59 Bits empfangen wurden
}


// --------- Zeitabgleich --------

void dcf77_exec (void) {

	if (sicher_count == 6){			//nur bei geprfter bertragung wird aktualisiert
		dcf77_sek = 0;
		dcf77_min = dcf77_min_h;
		dcf77_std = dcf77_std_h;
		dcf77_tag = dcf77_tag_h;
		dcf77_mon = dcf77_mon_h;
		dcf77_jah = dcf77_jah_h;
	}

	//zurcksetzten der empfangenen Zeit
	dcf77_min_h  = 0;
	dcf77_std_h  = 0;
	dcf77_tag_h  = 0;
	dcf77_mon_h  = 0;
	dcf77_jah_h  = 0;
	sicher_count = 0;
	parity = 0; 
}


// --------- Addiert eine Sekunde --------

void add_sek (void) {
	dcf77_sek++;

	if (dcf77_sek == 60){
		dcf77_sek = 0;
		dcf77_min++;

		if (dcf77_min == 60){
			dcf77_min = 0;
			dcf77_std++;

			if (dcf77_std == 24){
				dcf77_std = 0;
				dcf77_tag++;
			}
		}
	}
}


// --------- Zeit als String ausgeben --------

char *dcf77_zeit(void){
	char  stunde[2]= {'\0'};
	char  minute[2]= {'\0'};
	char sekunde[2]= {'\0'};
	int i;

	for(i=0; i<=10; i++) dcf77_zeitausg[i] = '\0';

	if(dcf77_std <= 9) sprintf(stunde, "0%d", dcf77_std);
	else sprintf(stunde, "%d", dcf77_std);

	if(dcf77_min <= 9) sprintf(minute, "0%d", dcf77_min);
	else sprintf(minute, "%d", dcf77_min);

	if(dcf77_sek <= 9) sprintf(sekunde, "0%d", dcf77_sek);
	else sprintf(sekunde, "%d", dcf77_sek);

	dcf77_zeitausg[0] = stunde[0];
	dcf77_zeitausg[1] = stunde[1];
	dcf77_zeitausg[2] = ':';
	dcf77_zeitausg[3] = minute[0];
	dcf77_zeitausg[4] = minute[1];
	dcf77_zeitausg[5] = ':';
	dcf77_zeitausg[6] = sekunde[0];
	dcf77_zeitausg[7] = sekunde[1];
	dcf77_zeitausg[8] = '\0';

	return dcf77_zeitausg;
}


// --------- Datum als String ausgeben --------

char *dcf77_datum(void){
	char   tag[2]= {'\0'};
	char monat[2]= {'\0'};
	char  jahr[2]= {'\0'};
	int i;

	for(i=0; i<=10; i++) dcf77_datumausg[i] = '\0';

	if(dcf77_tag <= 9) sprintf(tag, "0%d", dcf77_tag);
	else sprintf(tag, "%d", dcf77_tag);

	if(dcf77_mon <= 9) sprintf(monat, "0%d", dcf77_mon);
	else sprintf(monat, "%d", dcf77_mon);

	if(dcf77_jah <= 9) sprintf(jahr, "0%d", dcf77_jah);
	else sprintf(jahr, "%d", dcf77_jah);

	dcf77_datumausg[0] = tag[0];
	dcf77_datumausg[1] = tag[1];
	dcf77_datumausg[2] = '/';
	dcf77_datumausg[3] = monat[0];
	dcf77_datumausg[4] = monat[1];
	dcf77_datumausg[5] = '/';
	dcf77_datumausg[6] = jahr[0];
	dcf77_datumausg[7] = jahr[1];
	dcf77_datumausg[8] = '\0';

	return dcf77_datumausg;
}


